"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

import logging
from typing import Dict

from gravitino.api.catalog import Catalog
from gravitino.api.schema import Schema
from gravitino.api.schema_change import SchemaChange
from gravitino.api.supports_schemas import SupportsSchemas
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.catalog_dto import CatalogDTO
from gravitino.dto.requests.schema_create_request import SchemaCreateRequest
from gravitino.dto.requests.schema_update_request import SchemaUpdateRequest
from gravitino.dto.requests.schema_updates_request import SchemaUpdatesRequest
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.dto.responses.schema_response import SchemaResponse
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace
from gravitino.utils import HTTPClient

logger = logging.getLogger(__name__)


class BaseSchemaCatalog(CatalogDTO, SupportsSchemas):
    """
    BaseSchemaCatalog is the base abstract class for all the catalog with schema. It provides the
    common methods for managing schemas in a catalog. With BaseSchemaCatalog, users can list,
    create, load, alter and drop a schema with specified identifier.
    """

    rest_client: HTTPClient
    """The REST client to send the requests."""

    def __init__(
        self,
        name: str = None,
        catalog_type: Catalog.Type = Catalog.Type.UNSUPPORTED,
        provider: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
        audit: AuditDTO = None,
        rest_client: HTTPClient = None,
    ):
        super().__init__(
            _name=name,
            _type=catalog_type,
            _provider=provider,
            _comment=comment,
            _properties=properties,
            _audit=audit,
        )
        self.rest_client = rest_client

    def as_schemas(self):
        return self

    def list_schemas(self, namespace: Namespace) -> [NameIdentifier]:
        """List all the schemas under the given catalog namespace.

        Args:
             namespace: The namespace of the catalog.

        Raises:
            NoSuchCatalogException if the catalog with specified namespace does not exist.

        Returns:
             A list of {@link NameIdentifier} of the schemas under the given catalog namespace.
        """
        Namespace.check_schema(namespace)
        resp = self.rest_client.get(
            BaseSchemaCatalog.format_schema_request_path(namespace)
        )
        entity_list_response = EntityListResponse.from_json(
            resp.body, infer_missing=True
        )
        entity_list_response.validate()
        return entity_list_response.identifiers()

    def create_schema(
        self,
        ident: NameIdentifier = None,
        comment: str = None,
        properties: Dict[str, str] = None,
    ) -> Schema:
        """Create a new schema with specified identifier, comment and metadata.

        Args:
            ident: The name identifier of the schema.
            comment: The comment of the schema.
            properties: The properties of the schema.

        Raises:
            NoSuchCatalogException if the catalog with specified namespace does not exist.
            SchemaAlreadyExistsException if the schema with specified identifier already exists.

        Returns:
             The created Schema.
        """
        NameIdentifier.check_schema(ident)
        req = SchemaCreateRequest(ident.name(), comment, properties)
        req.validate()

        resp = self.rest_client.post(
            BaseSchemaCatalog.format_schema_request_path(ident.namespace()), json=req
        )
        schema_response = SchemaResponse.from_json(resp.body, infer_missing=True)
        schema_response.validate()

        return schema_response.schema()

    def load_schema(self, ident: NameIdentifier) -> Schema:
        """Load the schema with specified identifier.

        Args:
            ident: The name identifier of the schema.

        Raises:
            NoSuchSchemaException if the schema with specified identifier does not exist.

        Returns:
            The Schema with specified identifier.
        """
        NameIdentifier.check_schema(ident)
        resp = self.rest_client.get(
            BaseSchemaCatalog.format_schema_request_path(ident.namespace())
            + "/"
            + ident.name()
        )
        schema_response = SchemaResponse.from_json(resp.body, infer_missing=True)
        schema_response.validate()

        return schema_response.schema()

    def alter_schema(self, ident: NameIdentifier, *changes: SchemaChange) -> Schema:
        """Alter the schema with specified identifier by applying the changes.

        Args:
             ident: The name identifier of the schema.
            changes: The metadata changes to apply.

        Raises:
            NoSuchSchemaException if the schema with specified identifier does not exist.

        Returns:
            The altered Schema.
        """
        NameIdentifier.check_schema(ident)
        reqs = [
            BaseSchemaCatalog.to_schema_update_request(change) for change in changes
        ]
        updates_request = SchemaUpdatesRequest(reqs)
        updates_request.validate()
        resp = self.rest_client.put(
            BaseSchemaCatalog.format_schema_request_path(ident.namespace())
            + "/"
            + ident.name(),
            updates_request,
        )
        schema_response = SchemaResponse.from_json(resp.body, infer_missing=True)
        schema_response.validate()
        return schema_response.schema()

    def drop_schema(self, ident: NameIdentifier, cascade: bool) -> bool:
        """Drop the schema with specified identifier.

        Args:
            ident: The name identifier of the schema.
            cascade: Whether to drop all the tables under the schema.

        Raises:
            NonEmptySchemaException if the schema is not empty and cascade is false.

        Returns:
             true if the schema is dropped successfully, false otherwise.
        """
        NameIdentifier.check_schema(ident)
        try:
            params = {"cascade": str(cascade)}
            resp = self.rest_client.delete(
                BaseSchemaCatalog.format_schema_request_path(ident.namespace())
                + "/"
                + ident.name(),
                params=params,
            )
            drop_resp = DropResponse.from_json(resp.body, infer_missing=True)
            drop_resp.validate()
            return drop_resp.dropped()
        except Exception:
            logger.warning("Failed to drop schema %s", ident)
            return False

    @staticmethod
    def format_schema_request_path(ns: Namespace):
        return "api/metalakes/" + ns.level(0) + "/catalogs/" + ns.level(1) + "/schemas"

    @staticmethod
    def to_schema_update_request(change: SchemaChange):
        if isinstance(change, SchemaChange.SetProperty):
            return SchemaUpdateRequest.SetSchemaPropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, SchemaChange.RemoveProperty):
            return SchemaUpdateRequest.RemoveSchemaPropertyRequest(change.property())
        raise ValueError(f"Unknown change type: {type(change).__name__}")



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from abc import abstractmethod
from typing import Optional, Dict

from gravitino.api.auditable import Auditable


class Schema(Auditable):
    """
    An interface representing a schema in the Catalog. A Schema is a
    basic container of relational objects, like tables, views, etc. A Schema can be self-nested,
    which means it can be schema1.schema2.table.

    This defines the basic properties of a schema. A catalog implementation with SupportsSchemas
    should implement this interface.
    """

    @abstractmethod
    def name(self) -> str:
        """Returns the name of the Schema."""
        pass

    def comment(self) -> Optional[str]:
        """Returns the comment of the Schema. None is returned if the comment is not set."""
        return None

    def properties(self) -> Dict[str, str]:
        """Returns the properties of the Schema. An empty dictionary is returned if no properties are set."""
        return {}



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

import base64
import os

from .auth_constants import AuthConstants
from .auth_data_provider import AuthDataProvider


class SimpleAuthProvider(AuthDataProvider):
    """SimpleAuthProvider will use the environment variable `GRAVITINO_USER` or
    the user of the system to generate a basic token for every request.

    """

    _token: bytes

    def __init__(self):
        gravitino_user = os.environ.get("GRAVITINO_USER")
        if gravitino_user is None or len(gravitino_user) == 0:
            gravitino_user = os.environ.get("user.name")

        if gravitino_user is None or len(gravitino_user) == 0:
            gravitino_user = "anonymous"

        user_information = f"{gravitino_user}:dummy"
        self._token = (
            AuthConstants.AUTHORIZATION_BASIC_HEADER
            + base64.b64encode(user_information.encode("utf-8")).decode("utf-8")
        ).encode("utf-8")

    def has_token_data(self) -> bool:
        return True

    def get_token_data(self) -> bytes:
        return self._token

    def close(self):
        pass



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from abc import abstractmethod
from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.api.fileset_change import FilesetChange
from gravitino.rest.rest_message import RESTRequest


@dataclass
class FilesetUpdateRequestBase(RESTRequest):
    _type: str = field(metadata=config(field_name="@type"))

    def __init__(self, action_type: str):
        self._type = action_type

    @abstractmethod
    def fileset_change(self):
        pass


class FilesetUpdateRequest:
    """Request to update a fileset."""

    @dataclass
    class RenameFilesetRequest(FilesetUpdateRequestBase):
        """The fileset update request for renaming a fileset."""

        _new_name: str = field(metadata=config(field_name="newName"))
        """The new name for the Fileset."""

        def __init__(self, new_name: str):
            super().__init__("rename")
            self._new_name = new_name

        def validate(self):
            """Validates the fields of the request.

            Raises:
                IllegalArgumentException if the new name is not set.
            """
            if not self._new_name:
                raise ValueError('"new_name" field is required and cannot be empty')

        def fileset_change(self):
            """Returns the fileset change.

            Returns:
                the fileset change.
            """
            return FilesetChange.rename(self._new_name)

    @dataclass
    class UpdateFilesetCommentRequest(FilesetUpdateRequestBase):
        """Represents a request to update the comment on a Fileset."""

        _new_comment: str = field(metadata=config(field_name="newComment"))
        """The new comment for the Fileset."""

        def __init__(self, new_comment: str):
            super().__init__("updateComment")
            self._new_comment = new_comment

        def validate(self):
            """Validates the fields of the request.

            Raises:
                IllegalArgumentException if the new comment is not set.
            """
            if not self._new_comment:
                raise ValueError('"new_comment" field is required and cannot be empty')

        def fileset_change(self):
            """Returns the fileset change"""
            return FilesetChange.update_comment(self._new_comment)

    @dataclass
    class SetFilesetPropertyRequest(FilesetUpdateRequestBase):
        """Represents a request to set a property on a Fileset."""

        _property: str = field(metadata=config(field_name="property"))
        """The property to set."""

        _value: str = field(metadata=config(field_name="value"))
        """The value of the property."""

        def __init__(self, fileset_property: str, value: str):
            super().__init__("setProperty")
            self._property = fileset_property
            self._value = value

        def validate(self):
            """Validates the fields of the request.

            Raises:
                 IllegalArgumentException if fileset_property or value are not set.
            """
            if not self._property:
                raise ValueError(
                    '"fileset_property" field is required and cannot be empty'
                )
            if not self._value:
                raise ValueError('"value" field is required and cannot be empty')

        def fileset_change(self):
            return FilesetChange.set_property(self._property, self._value)

    @dataclass
    class RemoveFilesetPropertyRequest(FilesetUpdateRequestBase):
        """Represents a request to remove a property from a Fileset."""

        _property: str = field(metadata=config(field_name="property"))
        """The property to remove."""

        def __init__(self, fileset_property: str):
            super().__init__("removeProperty")
            self._property = fileset_property

        def validate(self):
            """Validates the fields of the request.

            Raises:
                 IllegalArgumentException if fileset_property is not set.
            """
            if not self._property:
                raise ValueError(
                    '"fileset_property" field is required and cannot be empty'
                )

        def fileset_change(self):
            return FilesetChange.remove_property(self._property)



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from gravitino.api.catalog import Catalog
from gravitino.api.catalog_change import CatalogChange
from gravitino.catalog.fileset_catalog import FilesetCatalog
from gravitino.dto.catalog_dto import CatalogDTO
from gravitino.dto.requests.catalog_update_request import CatalogUpdateRequest
from gravitino.dto.requests.metalake_update_request import MetalakeUpdateRequest
from gravitino.api.metalake_change import MetalakeChange
from gravitino.utils import HTTPClient


class DTOConverters:
    """Utility class for converting between DTOs and domain objects."""

    @staticmethod
    def to_metalake_update_request(change: MetalakeChange) -> object:
        # Assuming MetalakeUpdateRequest has similar nested class structure for requests
        if isinstance(change, MetalakeChange.RenameMetalake):
            return MetalakeUpdateRequest.RenameMetalakeRequest(change.new_name())
        if isinstance(change, MetalakeChange.UpdateMetalakeComment):
            return MetalakeUpdateRequest.UpdateMetalakeCommentRequest(
                change.new_comment()
            )
        if isinstance(change, MetalakeChange.SetProperty):
            return MetalakeUpdateRequest.SetMetalakePropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, MetalakeChange.RemoveProperty):
            return MetalakeUpdateRequest.RemoveMetalakePropertyRequest(
                change.property()
            )

        raise ValueError(f"Unknown change type: {type(change).__name__}")

    @staticmethod
    def to_catalog(catalog: CatalogDTO, client: HTTPClient):
        if catalog.type() == Catalog.Type.FILESET:
            return FilesetCatalog(
                name=catalog.name(),
                catalog_type=catalog.type(),
                provider=catalog.provider(),
                comment=catalog.comment(),
                properties=catalog.properties(),
                audit=catalog.audit_info(),
                rest_client=client,
            )

        raise NotImplementedError("Unsupported catalog type: " + str(catalog.type()))

    @staticmethod
    def to_catalog_update_request(change: CatalogChange):
        if isinstance(change, CatalogChange.RenameCatalog):
            return CatalogUpdateRequest.RenameCatalogRequest(change.new_name())
        if isinstance(change, CatalogChange.UpdateCatalogComment):
            return CatalogUpdateRequest.UpdateCatalogCommentRequest(
                change.new_comment()
            )
        if isinstance(change, CatalogChange.SetProperty):
            # TODO
            # pylint: disable=too-many-function-args
            return CatalogUpdateRequest.SetCatalogPropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, CatalogChange.RemoveProperty):
            return CatalogUpdateRequest.RemoveCatalogPropertyRequest(change.property())

        raise ValueError(f"Unknown change type: {type(change).__name__}")



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from dataclasses import dataclass, field
from typing import Optional, Dict

from dataclasses_json import config, DataClassJsonMixin

from gravitino.api.fileset import Fileset
from gravitino.dto.audit_dto import AuditDTO


@dataclass
class FilesetDTO(Fileset, DataClassJsonMixin):
    """Represents a Fileset DTO (Data Transfer Object)."""

    _name: str = field(metadata=config(field_name="name"))
    _comment: Optional[str] = field(metadata=config(field_name="comment"))
    _type: Fileset.Type = field(metadata=config(field_name="type"))
    _properties: Dict[str, str] = field(metadata=config(field_name="properties"))
    _storage_location: str = field(
        default=None, metadata=config(field_name="storageLocation")
    )
    _audit: AuditDTO = field(default=None, metadata=config(field_name="audit"))

    def name(self) -> str:
        return self._name

    def type(self) -> Fileset.Type:
        return self._type

    def storage_location(self) -> str:
        return self._storage_location

    def comment(self) -> Optional[str]:
        return self._comment

    def properties(self) -> Dict[str, str]:
        return self._properties

    def audit_info(self) -> AuditDTO:
        return self._audit



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from typing import List, Dict

from gravitino.api.catalog import Catalog
from gravitino.api.catalog_change import CatalogChange
from gravitino.auth.auth_data_provider import AuthDataProvider
from gravitino.client.gravitino_client_base import GravitinoClientBase
from gravitino.client.gravitino_metalake import GravitinoMetalake


class NoSuchMetalakeException(Exception):
    pass


class NoSuchCatalogException(Exception):
    pass


class CatalogAlreadyExistsException(Exception):
    pass


class GravitinoClient(GravitinoClientBase):
    """Gravitino Client for a user to interact with the Gravitino API, allowing the client to list,
    load, create, and alter Catalog.

    It uses an underlying {@link RESTClient} to send HTTP requests and receive responses from the API.
    """

    _metalake: GravitinoMetalake

    def __init__(
        self,
        uri: str,
        metalake_name: str,
        check_version: bool = True,
        auth_data_provider: AuthDataProvider = None,
    ):
        """Constructs a new GravitinoClient with the given URI, authenticator and AuthDataProvider.

        Args:
            uri: The base URI for the Gravitino API.
            metalake_name: The specified metalake name.
            auth_data_provider: The provider of the data which is used for authentication.

        Raises:
            NoSuchMetalakeException if the metalake with specified name does not exist.
        """
        super().__init__(uri, check_version, auth_data_provider)
        self.check_metalake_name(metalake_name)
        self._metalake = super().load_metalake(metalake_name)

    def get_metalake(self) -> GravitinoMetalake:
        """Get the current metalake object

        Raises:
            NoSuchMetalakeException if the metalake with specified name does not exist.

        Returns:
            the GravitinoMetalake object
        """
        return self._metalake

    def list_catalogs(self) -> List[str]:
        return self.get_metalake().list_catalogs()

    def list_catalogs_info(self) -> List[Catalog]:
        return self.get_metalake().list_catalogs_info()

    def load_catalog(self, name: str) -> Catalog:
        return self.get_metalake().load_catalog(name)

    def create_catalog(
        self,
        name: str,
        catalog_type: Catalog.Type,
        provider: str,
        comment: str,
        properties: Dict[str, str],
    ) -> Catalog:
        return self.get_metalake().create_catalog(
            name, catalog_type, provider, comment, properties
        )

    def alter_catalog(self, name: str, *changes: CatalogChange):
        return self.get_metalake().alter_catalog(name, *changes)

    def drop_catalog(self, name: str):
        return self.get_metalake().drop_catalog(name)



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from dataclasses import dataclass, field
from typing import Optional, Dict

from dataclasses_json import DataClassJsonMixin, config

from gravitino.api.audit import Audit
from gravitino.dto.audit_dto import AuditDTO
from gravitino.api.metalake import Metalake


@dataclass
class MetalakeDTO(Metalake, DataClassJsonMixin):
    """Represents a Metalake Data Transfer Object (DTO) that implements the Metalake interface."""

    _name: str = field(metadata=config(field_name="name"))
    """The name of the Metalake DTO."""

    _comment: Optional[str] = field(metadata=config(field_name="comment"))
    """The comment of the Metalake DTO."""

    _properties: Optional[Dict[str, str]] = field(
        metadata=config(field_name="properties")
    )
    """The properties of the Metalake DTO."""

    _audit: Optional[AuditDTO] = field(metadata=config(field_name="audit"))
    """The audit information of the Metalake DTO."""

    def name(self) -> str:
        return self._name

    def comment(self) -> str:
        return self._comment

    def properties(self) -> Dict[str, str]:
        return self._properties

    def audit_info(self) -> Audit:
        return self._audit

    def equals(self, other):
        if self == other:
            return True
        if not isinstance(other, MetalakeDTO):
            return False
        return (
            self.name() == other.name()
            and self.comment() == other.comment()
            and self.property_equal(self.properties(), other.properties())
            and self.audit_info() == other.audit_info()
        )

    def property_equal(self, p1, p2):
        if p1 is None and p2 is None:
            return True
        if p1 is not None and not p1 and p2 is None:
            return True
        if p2 is not None and not p2 and p1 is None:
            return True
        return p1 == p2



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

import logging
from typing import List, Dict

from gravitino.api.catalog import Catalog
from gravitino.api.fileset import Fileset
from gravitino.api.fileset_change import FilesetChange
from gravitino.catalog.base_schema_catalog import BaseSchemaCatalog
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.requests.fileset_create_request import FilesetCreateRequest
from gravitino.dto.requests.fileset_update_request import FilesetUpdateRequest
from gravitino.dto.requests.fileset_updates_request import FilesetUpdatesRequest
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.dto.responses.fileset_response import FilesetResponse
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace
from gravitino.utils import HTTPClient

logger = logging.getLogger(__name__)


class FilesetCatalog(BaseSchemaCatalog):
    """
    Fileset catalog is a catalog implementation that supports fileset like metadata operations, for
    example, schemas and filesets list, creation, update and deletion. A Fileset catalog is under the metalake.
    """

    def __init__(
        self,
        name: str = None,
        catalog_type: Catalog.Type = Catalog.Type.UNSUPPORTED,
        provider: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
        audit: AuditDTO = None,
        rest_client: HTTPClient = None,
    ):

        super().__init__(
            name, catalog_type, provider, comment, properties, audit, rest_client
        )

    def as_fileset_catalog(self):
        return self

    def list_filesets(self, namespace: Namespace) -> List[NameIdentifier]:
        """List the filesets in a schema namespace from the catalog.

        Args:
            namespace A schema namespace.

        Raises:
            NoSuchSchemaException If the schema does not exist.

        Returns:
            An array of fileset identifiers in the namespace.
        """
        Namespace.check_fileset(namespace)

        resp = self.rest_client.get(self.format_fileset_request_path(namespace))
        entity_list_resp = EntityListResponse.from_json(resp.body, infer_missing=True)
        entity_list_resp.validate()

        return entity_list_resp.identifiers()

    def load_fileset(self, ident) -> Fileset:
        """Load fileset metadata by {@link NameIdentifier} from the catalog.

        Args:
            ident: A fileset identifier.

        Raises:
            NoSuchFilesetException If the fileset does not exist.

        Returns:
            The fileset metadata.
        """
        NameIdentifier.check_fileset(ident)

        resp = self.rest_client.get(
            f"{self.format_fileset_request_path(ident.namespace())}/{ident.name()}"
        )
        fileset_resp = FilesetResponse.from_json(resp.body, infer_missing=True)
        fileset_resp.validate()

        return fileset_resp.fileset()

    def create_fileset(
        self,
        ident: NameIdentifier,
        comment: str,
        fileset_type: Fileset.Type,
        storage_location: str,
        properties: Dict[str, str],
    ) -> Fileset:
        """Create a fileset metadata in the catalog.

        If the type of the fileset object is "MANAGED", the underlying storageLocation can be null,
        and Gravitino will manage the storage location based on the location of the schema.

        If the type of the fileset object is "EXTERNAL", the underlying storageLocation must be set.

        Args:
            ident: A fileset identifier.
            comment: The comment of the fileset.
            fileset_type: The type of the fileset.
            storage_location: The storage location of the fileset.
            properties: The properties of the fileset.

        Raises:
            NoSuchSchemaException If the schema does not exist.
            FilesetAlreadyExistsException If the fileset already exists.

        Returns:
            The created fileset metadata
        """
        NameIdentifier.check_fileset(ident)

        req = FilesetCreateRequest(
            name=ident.name(),
            comment=comment,
            fileset_type=fileset_type,
            storage_location=storage_location,
            properties=properties,
        )

        resp = self.rest_client.post(
            self.format_fileset_request_path(ident.namespace()), req
        )
        fileset_resp = FilesetResponse.from_json(resp.body, infer_missing=True)
        fileset_resp.validate()

        return fileset_resp.fileset()

    def alter_fileset(self, ident, *changes) -> Fileset:
        """Update a fileset metadata in the catalog.

        Args:
            ident: A fileset identifier.
            changes: The changes to apply to the fileset.

        Args:
            IllegalArgumentException If the changes are invalid.
            NoSuchFilesetException If the fileset does not exist.

        Returns:
            The updated fileset metadata.
        """
        NameIdentifier.check_fileset(ident)

        updates = [
            FilesetCatalog.to_fileset_update_request(change) for change in changes
        ]
        req = FilesetUpdatesRequest(updates)
        req.validate()

        resp = self.rest_client.put(
            f"{self.format_fileset_request_path(ident.namespace())}/{ident.name()}", req
        )
        fileset_resp = FilesetResponse.from_json(resp.body, infer_missing=True)
        fileset_resp.validate()

        return fileset_resp.fileset()

    def drop_fileset(self, ident: NameIdentifier) -> bool:
        """Drop a fileset from the catalog.

        The underlying files will be deleted if this fileset type is managed, otherwise, only the
        metadata will be dropped.

        Args:
             ident: A fileset identifier.

        Returns:
             true If the fileset is dropped, false the fileset did not exist.
        """
        try:
            NameIdentifier.check_fileset(ident)

            resp = self.rest_client.delete(
                f"{self.format_fileset_request_path(ident.namespace())}/{ident.name()}",
            )
            drop_resp = DropResponse.from_json(resp.body, infer_missing=True)
            drop_resp.validate()

            return drop_resp.dropped()
        except Exception as e:
            logger.warning("Failed to drop fileset %s: %s", ident, e)
            return False

    @staticmethod
    def format_fileset_request_path(namespace: Namespace) -> str:
        schema_ns = Namespace.of(namespace.level(0), namespace.level(1))
        return f"{BaseSchemaCatalog.format_schema_request_path(schema_ns)}/{namespace.level(2)}/filesets"

    @staticmethod
    def to_fileset_update_request(change: FilesetChange):
        if isinstance(change, FilesetChange.RenameFileset):
            return FilesetUpdateRequest.RenameFilesetRequest(change.new_name())
        if isinstance(change, FilesetChange.UpdateFilesetComment):
            return FilesetUpdateRequest.UpdateFilesetCommentRequest(
                change.new_comment()
            )
        if isinstance(change, FilesetChange.SetProperty):
            return FilesetUpdateRequest.SetFilesetPropertyRequest(
                change.property(), change.value()
            )
        if isinstance(change, FilesetChange.RemoveProperty):
            return FilesetUpdateRequest.RemoveFilesetPropertyRequest(change.property())
        raise ValueError(f"Unknown change type: {type(change).__name__}")



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from abc import ABC, abstractmethod
from typing import List, Dict

from gravitino.api.schema import Schema
from gravitino.api.schema_change import SchemaChange
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace


class NoSuchSchemaException(Exception):
    """Exception raised if the schema does not exist."""

    pass


class SupportsSchemas(ABC):
    """
    The Catalog interface to support schema operations. If the implemented catalog has schema
    semantics, it should implement this interface.
    """

    @abstractmethod
    def list_schemas(self, namespace: Namespace) -> List[NameIdentifier]:
        """List schemas under a namespace.

        If an entity such as a table, view exists, its parent schemas must also exist and must be
        returned by this discovery method. For example, if table a.b.t exists, this method invoked as
        list_schemas(a) must return [a.b] in the result array.

        Args:
            namespace: The namespace to list.

        Raises:
            NoSuchCatalogException: If the catalog does not exist.

        Returns:
            A list of schema identifiers under the namespace.
        """
        pass

    def schema_exists(self, ident: NameIdentifier) -> bool:
        """Check if a schema exists.

        If an entity such as a table, view exists, its parent namespaces must also exist. For
        example, if table a.b.t exists, this method invoked as schema_exists(a.b) must return true.

        Args:
            ident: The name identifier of the schema.

        Returns:
            True if the schema exists, false otherwise.
        """
        try:
            self.load_schema(ident)
            return True
        except NoSuchSchemaException:
            return False

    @abstractmethod
    def create_schema(
        self, ident: NameIdentifier, comment: str, properties: Dict[str, str]
    ) -> Schema:
        """Create a schema in the catalog.

        Args:
            ident: The name identifier of the schema.
            comment: The comment of the schema.
            properties: The properties of the schema.

        Raises:
            NoSuchCatalogException: If the catalog does not exist.
            SchemaAlreadyExistsException: If the schema already exists.

        Returns:
            The created schema.
        """
        pass

    @abstractmethod
    def load_schema(self, ident: NameIdentifier) -> Schema:
        """Load metadata properties for a schema.

        Args:
            ident: The name identifier of the schema.

        Raises:
            NoSuchSchemaException: If the schema does not exist (optional).

        Returns:
            A schema.
        """
        pass

    @abstractmethod
    def alter_schema(self, ident: NameIdentifier, *changes: SchemaChange) -> Schema:
        """Apply the metadata change to a schema in the catalog.

        Args:
            ident: The name identifier of the schema.
            changes: The metadata changes to apply.

        Raises:
            NoSuchSchemaException: If the schema does not exist.

        Returns:
            The altered schema.
        """
        pass

    @abstractmethod
    def drop_schema(self, ident: NameIdentifier, cascade: bool) -> bool:
        """Drop a schema from the catalog. If cascade option is true, recursively
        drop all objects within the schema.

        Args:
            ident: The name identifier of the schema.
            cascade: If true, recursively drop all objects within the schema.

        Returns:
            True if the schema exists and is dropped successfully, false otherwise.

        Raises:
            NonEmptySchemaException: If the schema is not empty and cascade is false.
        """
        pass



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from typing import Dict
from dataclasses import dataclass, field

from dataclasses_json import config

from .audit_dto import AuditDTO
from ..api.catalog import Catalog


@dataclass
class CatalogDTO(Catalog):
    """Data transfer object representing catalog information."""

    _name: str = field(metadata=config(field_name="name"))
    _type: Catalog.Type = field(metadata=config(field_name="type"))
    _provider: str = field(metadata=config(field_name="provider"))
    _comment: str = field(metadata=config(field_name="comment"))
    _properties: Dict[str, str] = field(metadata=config(field_name="properties"))
    _audit: AuditDTO = field(default=None, metadata=config(field_name="audit"))

    def builder(
        self,
        name: str = None,
        catalog_type: Catalog.Type = Catalog.Type.UNSUPPORTED,
        provider: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
        audit: AuditDTO = None,
    ):
        self._name = name
        self._type = catalog_type
        self._provider = provider
        self._comment = comment
        self._properties = properties
        self._audit = audit

    def name(self) -> str:
        return self._name

    def type(self) -> Catalog.Type:
        return self._type

    def provider(self) -> str:
        return self._provider

    def comment(self) -> str:
        return self._comment

    def properties(self) -> Dict[str, str]:
        return self._properties

    def audit_info(self) -> AuditDTO:
        return self._audit



"""
MIT License

Copyright (c) 2016 Dhamu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json


class HTTPError(Exception):
    """Base of all other errors"""

    def __init__(self, error):
        self.status_code = error.code
        self.reason = error.reason
        self.body = error.read()
        self.headers = error.hdrs

    def json(self):
        """
        :Returns: object of response error from the API
        """
        try:
            return json.loads(self.body.decode("utf-8"))
        except json.decoder.JSONDecodeError:
            return {"exception": self.body.decode("utf-8")}

    def __str__(self):
        return self.json().get("exception")


class BadRequestsError(HTTPError):
    pass


class UnauthorizedError(HTTPError):
    pass


class ForbiddenError(HTTPError):
    pass


class NotFoundError(HTTPError):
    pass


class MethodNotAllowedError(HTTPError):
    pass


class PayloadTooLargeError(HTTPError):
    pass


class UnsupportedMediaTypeError(HTTPError):
    pass


class TooManyRequestsError(HTTPError):
    pass


class InternalServerError(HTTPError):
    pass


class ServiceUnavailableError(HTTPError):
    pass


class GatewayTimeoutError(HTTPError):
    pass


err_dict = {
    400: BadRequestsError,
    401: UnauthorizedError,
    403: ForbiddenError,
    404: NotFoundError,
    405: MethodNotAllowedError,
    413: PayloadTooLargeError,
    415: UnsupportedMediaTypeError,
    429: TooManyRequestsError,
    500: InternalServerError,
    503: ServiceUnavailableError,
    504: GatewayTimeoutError,
}


def handle_error(error):
    try:
        exc = err_dict[error.code](error)
    except KeyError:
        return HTTPError(error)
    return exc



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

import logging
from typing import List, Dict

from gravitino.client.gravitino_client_base import GravitinoClientBase
from gravitino.client.gravitino_metalake import GravitinoMetalake
from gravitino.dto.dto_converters import DTOConverters
from gravitino.dto.requests.metalake_create_request import MetalakeCreateRequest
from gravitino.dto.requests.metalake_updates_request import MetalakeUpdatesRequest
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.metalake_list_response import MetalakeListResponse
from gravitino.dto.responses.metalake_response import MetalakeResponse
from gravitino.api.metalake_change import MetalakeChange

logger = logging.getLogger(__name__)


class GravitinoAdminClient(GravitinoClientBase):
    """
    Gravitino Client for the administrator to interact with the Gravitino API.
    It allows the client to list, load, create, and alter Metalakes.
    Normal users should use {@link GravitinoClient} to connect with the Gravitino server.
    """

    def list_metalakes(self) -> List[GravitinoMetalake]:
        """Retrieves a list of Metalakes from the Gravitino API.

        Returns:
            An array of GravitinoMetalake objects representing the Metalakes.
        """
        resp = self._rest_client.get(self.API_METALAKES_LIST_PATH)
        metalake_list_resp = MetalakeListResponse.from_json(
            resp.body, infer_missing=True
        )
        metalake_list_resp.validate()

        return [
            GravitinoMetalake(o, self._rest_client)
            for o in metalake_list_resp.metalakes()
        ]

    def create_metalake(
        self, name: str, comment: str, properties: Dict[str, str]
    ) -> GravitinoMetalake:
        """Creates a new Metalake using the Gravitino API.

        Args:
            name: The name of the new Metalake.
            comment: The comment for the new Metalake.
            properties: The properties of the new Metalake.

        Returns:
            A GravitinoMetalake instance representing the newly created Metalake.
            TODO: @throws MetalakeAlreadyExistsException If a Metalake with the specified identifier already exists.
        """
        req = MetalakeCreateRequest(name, comment, properties)
        req.validate()

        resp = self._rest_client.post(self.API_METALAKES_LIST_PATH, req)
        metalake_response = MetalakeResponse.from_json(resp.body, infer_missing=True)
        metalake_response.validate()
        metalake = metalake_response.metalake()

        return GravitinoMetalake(metalake, self._rest_client)

    def alter_metalake(self, name: str, *changes: MetalakeChange) -> GravitinoMetalake:
        """Alters a specific Metalake using the Gravitino API.

        Args:
            name: The name of the Metalake to be altered.
            changes: The changes to be applied to the Metalake.

        Returns:
             A GravitinoMetalake instance representing the updated Metalake.
        TODO: @throws NoSuchMetalakeException If the specified Metalake does not exist.
        TODO: @throws IllegalArgumentException If the provided changes are invalid or not applicable.
        """

        reqs = [DTOConverters.to_metalake_update_request(change) for change in changes]
        updates_request = MetalakeUpdatesRequest(reqs)
        updates_request.validate()

        resp = self._rest_client.put(
            self.API_METALAKES_IDENTIFIER_PATH + name, updates_request
        )
        metalake_response = MetalakeResponse.from_json(resp.body, infer_missing=True)
        metalake_response.validate()
        metalake = metalake_response.metalake()

        return GravitinoMetalake(metalake, self._rest_client)

    def drop_metalake(self, name: str) -> bool:
        """Drops a specific Metalake using the Gravitino API.

        Args:
            name: The name of the Metalake to be dropped.

        Returns:
            True if the Metalake was successfully dropped, false otherwise.
        """
        try:
            resp = self._rest_client.delete(self.API_METALAKES_IDENTIFIER_PATH + name)
            drop_response = DropResponse.from_json(resp.body, infer_missing=True)

            return drop_response.dropped()
        except Exception:
            logger.warning("Failed to drop metalake %s", name)
            return False



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from typing import List, ClassVar


class Namespace:
    """A namespace is a sequence of levels separated by dots. It's used to identify a metalake, a
    catalog or a schema. For example, "metalake1", "metalake1.catalog1" and
    "metalake1.catalog1.schema1" are all valid namespaces.
    """

    _DOT: ClassVar[str] = "."

    _levels: List[str] = []

    def __init__(self, levels: List[str]):
        self._levels = levels

    @staticmethod
    def empty() -> "Namespace":
        """Get an empty namespace.

        Returns:
            An empty namespace
        """
        return Namespace([])

    @staticmethod
    def of(*levels: str) -> "Namespace":
        """Create a namespace with the given levels.

        Args:
            levels The levels of the namespace

        Returns:
            A namespace with the given levels
        """
        Namespace.check(
            levels is not None, "Cannot create a namespace with null levels"
        )
        if len(levels) == 0:
            return Namespace.empty()

        for level in levels:
            Namespace.check(
                level is not None and level != "",
                "Cannot create a namespace with null or empty level",
            )

        return Namespace(list(levels))

    @staticmethod
    def of_metalake() -> "Namespace":
        """Create a namespace for metalake.

        Returns:
            A namespace for metalake
        """
        return Namespace.empty()

    @staticmethod
    def of_catalog(metalake: str) -> "Namespace":
        """Create a namespace for catalog.

        Args:
            metalake: The metalake name

        Returns:
            A namespace for catalog
        """
        return Namespace.of(metalake)

    @staticmethod
    def of_schema(metalake: str, catalog: str) -> "Namespace":
        """Create a namespace for schema.

        Args:
            metalake: The metalake name
            catalog: The catalog name

        Returns:
             A namespace for schema
        """
        return Namespace.of(metalake, catalog)

    @staticmethod
    def of_table(metalake: str, catalog: str, schema: str) -> "Namespace":
        """Create a namespace for table.

        Args:
            metalake: The metalake name
            catalog: The catalog name
            schema: The schema name

        Returns:
             A namespace for table
        """
        return Namespace.of(metalake, catalog, schema)

    @staticmethod
    def of_fileset(metalake: str, catalog: str, schema: str) -> "Namespace":
        """Create a namespace for fileset.

        Args:
            metalake: The metalake name
            catalog: The catalog name
            schema: The schema name

        Returns:
             A namespace for fileset
        """
        return Namespace.of(metalake, catalog, schema)

    @staticmethod
    def of_topic(metalake: str, catalog: str, schema: str) -> "Namespace":
        """Create a namespace for topic.

        Args:
            metalake: The metalake name
            catalog: The catalog name
            schema: The schema name

        Returns:
             A namespace for topic
        """
        return Namespace.of(metalake, catalog, schema)

    @staticmethod
    def check_metalake(namespace: "Namespace") -> None:
        """Check if the given metalake namespace is legal, throw an IllegalNamespaceException if
        it's illegal.

        Args:
            namespace: The metalake namespace
        """
        Namespace.check(
            namespace is not None and namespace.is_empty(),
            f"Metalake namespace must be non-null and empty, the input namespace is {namespace}",
        )

    @staticmethod
    def check_catalog(namespace: "Namespace") -> None:
        """Check if the given catalog namespace is legal, throw an IllegalNamespaceException if
        it's illegal.

        Args:
            namespace: The catalog namespace
        """
        Namespace.check(
            namespace is not None and namespace.length() == 1,
            f"Catalog namespace must be non-null and have 1 level, the input namespace is {namespace}",
        )

    @staticmethod
    def check_schema(namespace: "Namespace") -> None:
        """Check if the given schema namespace is legal, throw an IllegalNamespaceException if
        it's illegal.

        Args:
            namespace: The schema namespace
        """
        Namespace.check(
            namespace is not None and namespace.length() == 2,
            f"Schema namespace must be non-null and have 2 levels, the input namespace is {namespace}",
        )

    @staticmethod
    def check_table(namespace: "Namespace") -> None:
        """Check if the given table namespace is legal, throw an IllegalNamespaceException if it's
        illegal.

        Args:
            namespace: The table namespace
        """
        Namespace.check(
            namespace is not None and namespace.length() == 3,
            f"Table namespace must be non-null and have 3 levels, the input namespace is {namespace}",
        )

    @staticmethod
    def check_fileset(namespace: "Namespace") -> None:
        """Check if the given fileset namespace is legal, throw an IllegalNamespaceException if
        it's illegal.

        Args:
            namespace: The fileset namespace
        """
        Namespace.check(
            namespace is not None and namespace.length() == 3,
            f"Fileset namespace must be non-null and have 3 levels, the input namespace is {namespace}",
        )

    @staticmethod
    def check_topic(namespace: "Namespace") -> None:
        """Check if the given topic namespace is legal, throw an IllegalNamespaceException if it's
        illegal.

        Args:
            namespace: The topic namespace
        """
        Namespace.check(
            namespace is not None and namespace.length() == 3,
            f"Topic namespace must be non-null and have 3 levels, the input namespace is {namespace}",
        )

    def levels(self) -> List[str]:
        """Get the levels of the namespace.

        Returns:
            The levels of the namespace
        """
        return self._levels

    def level(self, pos: int) -> str:
        """Get the level at the given position.

        Args:
            pos: The position of the level

        Returns:
            The level at the given position
        """
        if pos < 0 or pos >= len(self._levels):
            raise ValueError("Invalid level position")
        return self._levels[pos]

    def length(self) -> int:
        """Get the length of the namespace.

        Returns:
            The length of the namespace.
        """
        return len(self._levels)

    def is_empty(self) -> bool:
        """Check if the namespace is empty.

        Returns:
            True if the namespace is empty, false otherwise.
        """
        return len(self._levels) == 0

    def __eq__(self, other: "Namespace") -> bool:
        if not isinstance(other, Namespace):
            return False
        return self._levels == other._levels

    def __hash__(self) -> int:
        return hash(tuple(self._levels))

    def __str__(self) -> str:
        return self._DOT.join(self._levels)

    @staticmethod
    def check(expression: bool, message: str, *args) -> None:
        """Check the given condition is true. Throw an IllegalNamespaceException if it's not.

        Args:
            expression: The expression to check.
            message: The message to throw.
            args: The arguments to the message.
        """
        if not expression:
            raise ValueError(message.format(*args))



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from dataclasses import dataclass, field

from dataclasses_json import config


class MetalakeChange:
    """A metalake change is a change to a metalake. It can be used to rename a metalake, update the
    comment of a metalake, set a property and value pair for a metalake, or remove a property from a
    metalake.
    """

    @staticmethod
    def rename(new_name: str) -> "MetalakeChange.RenameMetalake":
        """Creates a new metalake change to rename the metalake.

        Args:
            new_name: The New name of the metalake.

        Returns:
            The metalake change.
        """
        return MetalakeChange.RenameMetalake(new_name)

    @staticmethod
    def update_comment(new_comment: str) -> "MetalakeChange.UpdateMetalakeComment":
        """Creates a new metalake change to update the metalake comment.

        Args:
            new_comment: The new comment of the metalake.

        Returns:
            The metalake change.
        """
        return MetalakeChange.UpdateMetalakeComment(new_comment)

    @staticmethod
    def set_property(metalake_property: str, value: str) -> "SetProperty":
        """Creates a new metalake change to set a property and value pair for the metalake.

        Args:
            metalake_property: The property name to set.
            value: The value to set the property to.

        Returns:
             The metalake change.
        """
        return MetalakeChange.SetProperty(metalake_property, value)

    @staticmethod
    def remove_property(metalake_property: str) -> "RemoveProperty":
        """Creates a new metalake change to remove a property from the metalake.

        Args:
            metalake_property: The property name to remove.

        Returns:
            The metalake change.
        """
        return MetalakeChange.RemoveProperty(metalake_property)

    @dataclass(frozen=True)
    class RenameMetalake:
        """A metalake change to rename the metalake."""

        _new_name: str = field(metadata=config(field_name="new_name"))

        def new_name(self) -> str:
            return self._new_name

        def __str__(self):
            return f"RENAMEMETALAKE {self._new_name}"

    @dataclass(frozen=True)
    class UpdateMetalakeComment:
        """A metalake change to update the metalake comment"""

        _new_comment: str = field(metadata=config(field_name="new_comment"))

        def new_comment(self) -> str:
            return self._new_comment

        def __str__(self):
            return f"UPDATEMETALAKECOMMENT {self._new_comment}"

    @dataclass(frozen=True)
    class SetProperty:
        """A metalake change to set a property and value pair for the metalake"""

        _property: str = field(metadata=config(field_name="property"))
        _value: str = field(metadata=config(field_name="value"))

        def property(self) -> str:
            return self._property

        def value(self) -> str:
            return self._value

        def __str__(self):
            return f"SETPROPERTY {self._property} {self._value}"

    @dataclass(frozen=True)
    class RemoveProperty:
        """A metalake change to remove a property from the metalake"""

        _property: str

        def property(self) -> str:
            return self._property

        def __str__(self):
            return f"REMOVEPROPERTY {self.property}"



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from abc import abstractmethod
from typing import Optional, Dict

from gravitino.api.auditable import Auditable


class Metalake(Auditable):
    """
    The interface of a metalake. The metalake is the top level entity in the gravitino system,
    containing a set of catalogs.
    """

    @abstractmethod
    def name(self) -> str:
        """The name of the metalake.

        Returns:
            str: The name of the metalake.
        """
        pass

    @abstractmethod
    def comment(self) -> Optional[str]:
        """The comment of the metalake. Note. this method will return None if the comment is not set for
        this metalake.

        Returns:
            Optional[str]: The comment of the metalake.
        """
        pass

    @abstractmethod
    def properties(self) -> Optional[Dict[str, str]]:
        """The properties of the metalake. Note, this method will return None if the properties are not
        set.

        Returns:
            Optional[Dict[str, str]]: The properties of the metalake.
        """
        pass



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

"""

from enum import Enum
from pathlib import PurePosixPath
from typing import Dict, Tuple
import re
import fsspec

from cachetools import TTLCache
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem
from fsspec.implementations.arrow import ArrowFSWrapper
from fsspec.utils import infer_storage_options
from pyarrow.fs import HadoopFileSystem
from readerwriterlock import rwlock
from gravitino.api.catalog import Catalog
from gravitino.api.fileset import Fileset
from gravitino.client.gravitino_client import GravitinoClient
from gravitino.exceptions.gravitino_runtime_exception import GravitinoRuntimeException
from gravitino.name_identifier import NameIdentifier

PROTOCOL_NAME = "gvfs"


class StorageType(Enum):
    HDFS = "hdfs"
    LOCAL = "file"


class FilesetContext:
    """A context object that holds the information about the fileset and the file system which used in
    the GravitinoVirtualFileSystem's operations.
    """

    def __init__(
        self,
        name_identifier: NameIdentifier,
        fileset: Fileset,
        fs: AbstractFileSystem,
        storage_type: StorageType,
        actual_path: str,
    ):
        self._name_identifier = name_identifier
        self._fileset = fileset
        self._fs = fs
        self._storage_type = storage_type
        self._actual_path = actual_path

    def get_name_identifier(self):
        return self._name_identifier

    def get_fileset(self):
        return self._fileset

    def get_fs(self):
        return self._fs

    def get_actual_path(self):
        return self._actual_path

    def get_storage_type(self):
        return self._storage_type


class GravitinoVirtualFileSystem(fsspec.AbstractFileSystem):
    """This is a virtual file system which users can access `fileset` and
    other resources.

    It obtains the actual storage location corresponding to the resource from the
    Gravitino server, and creates an independent file system for it to act as an agent for users to
    access the underlying storage.
    """

    # Override the parent variable
    protocol = PROTOCOL_NAME
    _identifier_pattern = re.compile("^fileset/([^/]+)/([^/]+)/([^/]+)(?:/[^/]+)*/?$")

    def __init__(
        self,
        server_uri=None,
        metalake_name=None,
        cache_size=20,
        cache_expired_time=3600,
        **kwargs,
    ):
        self._metalake = metalake_name
        self._client = GravitinoClient(
            uri=server_uri, metalake_name=metalake_name, check_version=False
        )
        self._cache = TTLCache(maxsize=cache_size, ttl=cache_expired_time)
        self._cache_lock = rwlock.RWLockFair()

        super().__init__(**kwargs)

    @property
    def cache(self):
        return self._cache

    @property
    def fsid(self):
        return PROTOCOL_NAME

    def sign(self, path, expiration=None, **kwargs):
        """We do not support to create a signed URL representing the given path in gvfs."""
        raise GravitinoRuntimeException(
            "Sign is not implemented for Gravitino Virtual FileSystem."
        )

    def ls(self, path, detail=True, **kwargs):
        """List the files and directories info of the path.
        :param path: Virtual fileset path
        :param detail: Whether to show the details for the files and directories info
        :param kwargs: Extra args
        :return If details is true, returns a list of file info dicts, else returns a list of file paths
        """
        context: FilesetContext = self._get_fileset_context(path)
        if detail:
            entries = [
                self._convert_actual_info(entry, context)
                for entry in context.get_fs().ls(
                    self._strip_storage_protocol(
                        context.get_storage_type(), context.get_actual_path()
                    ),
                    detail=True,
                )
            ]
            return entries
        entries = [
            self._convert_actual_path(entry_path, context)
            for entry_path in context.get_fs().ls(
                self._strip_storage_protocol(
                    context.get_storage_type(), context.get_actual_path()
                ),
                detail=False,
            )
        ]
        return entries

    def info(self, path, **kwargs):
        """Get file info.
        :param path: Virtual fileset path
        :param kwargs: Extra args
        :return A file info dict
        """
        context: FilesetContext = self._get_fileset_context(path)
        actual_info: Dict = context.get_fs().info(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            )
        )
        return self._convert_actual_info(actual_info, context)

    def exists(self, path, **kwargs):
        """Check if a file or a directory exists.
        :param path: Virtual fileset path
        :param kwargs: Extra args
        :return If a file or directory exists, it returns True, otherwise False
        """
        context: FilesetContext = self._get_fileset_context(path)
        return context.get_fs().exists(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            )
        )

    def cp_file(self, path1, path2, **kwargs):
        """Copy a file.
        :param path1: Virtual src fileset path
        :param path2: Virtual dst fileset path, should be consistent with the src path fileset identifier
        :param kwargs: Extra args
        """
        src_path = self._pre_process_path(path1)
        dst_path = self._pre_process_path(path2)
        src_identifier: NameIdentifier = self._extract_identifier(src_path)
        dst_identifier: NameIdentifier = self._extract_identifier(dst_path)
        if src_identifier != dst_identifier:
            raise GravitinoRuntimeException(
                f"Destination file path identifier: `{dst_identifier}` should be same with src file path "
                f"identifier: `{src_identifier}`."
            )
        src_context: FilesetContext = self._get_fileset_context(src_path)
        if self._check_mount_single_file(
            src_context.get_fileset(),
            src_context.get_fs(),
            src_context.get_storage_type(),
        ):
            raise GravitinoRuntimeException(
                f"Cannot cp file of the fileset: {src_identifier} which only mounts to a single file."
            )
        dst_context: FilesetContext = self._get_fileset_context(dst_path)

        src_context.get_fs().cp_file(
            self._strip_storage_protocol(
                src_context.get_storage_type(), src_context.get_actual_path()
            ),
            self._strip_storage_protocol(
                dst_context.get_storage_type(), dst_context.get_actual_path()
            ),
        )

    def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
        """Move a file to another directory.
         This can move a file to another existing directory.
         If the target path directory does not exist, an exception will be thrown.
        :param path1: Virtual src fileset path
        :param path2: Virtual dst fileset path, should be consistent with the src path fileset identifier
        :param recursive: Whether to move recursively
        :param maxdepth: Maximum depth of recursive move
        :param kwargs: Extra args
        """
        src_path = self._pre_process_path(path1)
        dst_path = self._pre_process_path(path2)
        src_identifier: NameIdentifier = self._extract_identifier(src_path)
        dst_identifier: NameIdentifier = self._extract_identifier(dst_path)
        if src_identifier != dst_identifier:
            raise GravitinoRuntimeException(
                f"Destination file path identifier: `{dst_identifier}`"
                f" should be same with src file path identifier: `{src_identifier}`."
            )
        src_context: FilesetContext = self._get_fileset_context(src_path)
        if self._check_mount_single_file(
            src_context.get_fileset(),
            src_context.get_fs(),
            src_context.get_storage_type(),
        ):
            raise GravitinoRuntimeException(
                f"Cannot cp file of the fileset: {src_identifier} which only mounts to a single file."
            )
        dst_context: FilesetContext = self._get_fileset_context(dst_path)
        if src_context.get_storage_type() == StorageType.HDFS:
            src_context.get_fs().mv(
                self._strip_storage_protocol(
                    src_context.get_storage_type(), src_context.get_actual_path()
                ),
                self._strip_storage_protocol(
                    dst_context.get_storage_type(), dst_context.get_actual_path()
                ),
            )
        elif src_context.get_storage_type() == StorageType.LOCAL:
            src_context.get_fs().mv(
                self._strip_storage_protocol(
                    src_context.get_storage_type(), src_context.get_actual_path()
                ),
                self._strip_storage_protocol(
                    dst_context.get_storage_type(), dst_context.get_actual_path()
                ),
                recursive,
                maxdepth,
            )
        else:
            raise GravitinoRuntimeException(
                f"Storage type:{src_context.get_storage_type()} doesn't support now."
            )

    def _rm(self, path):
        raise GravitinoRuntimeException(
            "Deprecated method, use `rm_file` method instead."
        )

    def rm(self, path, recursive=False, maxdepth=None):
        """Remove a file or directory.
        :param path: Virtual fileset path
        :param recursive: Whether to remove the directory recursively.
                When removing a directory, this parameter should be True.
        :param maxdepth: The maximum depth to remove the directory recursively.
        """
        context: FilesetContext = self._get_fileset_context(path)
        context.get_fs().rm(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            ),
            recursive,
            maxdepth,
        )

    def rm_file(self, path):
        """Remove a file.
        :param path: Virtual fileset path
        """
        context: FilesetContext = self._get_fileset_context(path)
        context.get_fs().rm_file(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            )
        )

    def rmdir(self, path):
        """Remove a directory.
        It will delete a directory and all its contents recursively for PyArrow.HadoopFileSystem.
        And it will throw an exception if delete a directory which is non-empty for LocalFileSystem.
        :param path: Virtual fileset path
        """
        context: FilesetContext = self._get_fileset_context(path)
        context.get_fs().rmdir(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            )
        )

    def open(
        self,
        path,
        mode="rb",
        block_size=None,
        cache_options=None,
        compression=None,
        **kwargs,
    ):
        """Open a file to read/write/append.
        :param path: Virtual fileset path
        :param mode: The mode now supports: rb(read), wb(write), ab(append). See builtin ``open()``
        :param block_size: Some indication of buffering - this is a value in bytes
        :param cache_options: Extra arguments to pass through to the cache
        :param compression: If given, open file using compression codec
        :param kwargs: Extra args
        :return A file-like object from the filesystem
        """
        context: FilesetContext = self._get_fileset_context(path)
        return context.get_fs().open(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            ),
            mode,
            block_size,
            cache_options,
            compression,
            **kwargs,
        )

    def mkdir(self, path, create_parents=True, **kwargs):
        """Make a directory.
        if create_parents=True, this is equivalent to ``makedirs``.

        :param path: Virtual fileset path
        :param create_parents: Create parent directories if missing when set to True
        :param kwargs: Extra args
        """
        context: FilesetContext = self._get_fileset_context(path)
        context.get_fs().mkdir(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            ),
            create_parents,
            **kwargs,
        )

    def makedirs(self, path, exist_ok=True):
        """Make a directory recursively.
        :param path: Virtual fileset path
        :param exist_ok: Continue if a directory already exists
        """
        context: FilesetContext = self._get_fileset_context(path)
        context.get_fs().makedirs(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            ),
            exist_ok,
        )

    def created(self, path):
        """Return the created timestamp of a file as a datetime.datetime
        Only supports for `fsspec.LocalFileSystem` now.
        :param path: Virtual fileset path
        :return Created time(datetime.datetime)
        """
        context: FilesetContext = self._get_fileset_context(path)
        if context.get_storage_type() == StorageType.LOCAL:
            return context.get_fs().created(
                self._strip_storage_protocol(
                    context.get_storage_type(), context.get_actual_path()
                )
            )
        raise GravitinoRuntimeException(
            f"Storage type:{context.get_storage_type()} doesn't support now."
        )

    def modified(self, path):
        """Returns the modified time of the path file if it exists.
        :param path: Virtual fileset path
        :return Modified time(datetime.datetime)
        """
        context: FilesetContext = self._get_fileset_context(path)
        return context.get_fs().modified(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            )
        )

    def cat_file(self, path, start=None, end=None, **kwargs):
        """Get the content of a file.
        :param path: Virtual fileset path
        :param start: The offset in bytes to start reading from. It can be None.
        :param end: The offset in bytes to end reading at. It can be None.
        :param kwargs: Extra args
        :return File content
        """
        context: FilesetContext = self._get_fileset_context(path)
        return context.get_fs().cat_file(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            ),
            start,
            end,
            **kwargs,
        )

    def get_file(self, rpath, lpath, callback=None, outfile=None, **kwargs):
        """Copy single remote file to local.
        :param rpath: Remote file path
        :param lpath: Local file path
        :param callback: The callback class
        :param outfile: The output file path
        :param kwargs: Extra args
        """
        if not lpath.startswith(f"{StorageType.LOCAL.value}:") and not lpath.startswith(
            "/"
        ):
            raise GravitinoRuntimeException(
                "Doesn't support copy a remote gvfs file to an another remote file."
            )
        context: FilesetContext = self._get_fileset_context(rpath)
        context.get_fs().get_file(
            self._strip_storage_protocol(
                context.get_storage_type(), context.get_actual_path()
            ),
            lpath,
            **kwargs,
        )

    def _convert_actual_path(self, path, context: FilesetContext):
        """Convert an actual path to a virtual path.
          The virtual path is like `fileset/{catalog}/{schema}/{fileset}/xxx`.
        :param path: Actual path
        :param context: Fileset context
        :return A virtual path
        """
        if context.get_storage_type() == StorageType.HDFS:
            actual_prefix = infer_storage_options(
                context.get_fileset().storage_location()
            )["path"]
        elif context.get_storage_type() == StorageType.LOCAL:
            actual_prefix = context.get_fileset().storage_location()[
                len(f"{StorageType.LOCAL.value}:") :
            ]
        else:
            raise GravitinoRuntimeException(
                f"Storage type:{context.get_storage_type()} doesn't support now."
            )

        if not path.startswith(actual_prefix):
            raise GravitinoRuntimeException(
                f"Path {path} does not start with valid prefix {actual_prefix}."
            )
        virtual_location = self._get_virtual_location(context.get_name_identifier())
        return f"{path.replace(actual_prefix, virtual_location)}"

    def _convert_actual_info(self, entry: Dict, context: FilesetContext):
        """Convert a file info from an actual entry to a virtual entry.
        :param entry: A dict of the actual file info
        :param context: Fileset context
        :return A dict of the virtual file info
        """
        path = self._convert_actual_path(entry["name"], context)
        return {
            "name": path,
            "size": entry["size"],
            "type": entry["type"],
            "mtime": entry["mtime"],
        }

    def _get_fileset_context(self, virtual_path: str):
        """Get a fileset context from the cache or the Gravitino server
        :param virtual_path: The virtual path
        :return A fileset context
        """
        virtual_path: str = self._pre_process_path(virtual_path)
        identifier: NameIdentifier = self._extract_identifier(virtual_path)
        read_lock = self._cache_lock.gen_rlock()
        try:
            read_lock.acquire()
            cache_value: Tuple[Fileset, AbstractFileSystem, StorageType] = (
                self._cache.get(identifier)
            )
            if cache_value is not None:
                actual_path = self._get_actual_path_by_ident(
                    identifier,
                    cache_value[0],
                    cache_value[1],
                    cache_value[2],
                    virtual_path,
                )
                return FilesetContext(
                    identifier,
                    cache_value[0],
                    cache_value[1],
                    cache_value[2],
                    actual_path,
                )
        finally:
            read_lock.release()

        write_lock = self._cache_lock.gen_wlock()
        try:
            write_lock.acquire()
            cache_value: Tuple[Fileset, AbstractFileSystem] = self._cache.get(
                identifier
            )
            if cache_value is not None:
                actual_path = self._get_actual_path_by_ident(
                    identifier,
                    cache_value[0],
                    cache_value[1],
                    cache_value[2],
                    virtual_path,
                )
                return FilesetContext(
                    identifier,
                    cache_value[0],
                    cache_value[1],
                    cache_value[2],
                    actual_path,
                )
            fileset: Fileset = self._load_fileset_from_server(identifier)
            storage_location = fileset.storage_location()
            if storage_location.startswith(f"{StorageType.HDFS.value}://"):
                fs = ArrowFSWrapper(HadoopFileSystem.from_uri(storage_location))
                storage_type = StorageType.HDFS
            elif storage_location.startswith(f"{StorageType.LOCAL.value}:/"):
                fs = LocalFileSystem()
                storage_type = StorageType.LOCAL
            else:
                raise GravitinoRuntimeException(
                    f"Storage under the fileset: `{identifier}` doesn't support now."
                )
            actual_path = self._get_actual_path_by_ident(
                identifier, fileset, fs, storage_type, virtual_path
            )
            self._cache[identifier] = (fileset, fs, storage_type)
            context = FilesetContext(identifier, fileset, fs, storage_type, actual_path)
            return context
        finally:
            write_lock.release()

    def _extract_identifier(self, path):
        """Extract the fileset identifier from the path.
        :param path: The virtual fileset path
        :return The fileset identifier
        """
        if path is None:
            raise GravitinoRuntimeException(
                "path which need be extracted cannot be null or empty."
            )

        match = self._identifier_pattern.match(path)
        if match and len(match.groups()) == 3:
            return NameIdentifier.of_fileset(
                self._metalake, match.group(1), match.group(2), match.group(3)
            )
        raise GravitinoRuntimeException(
            f"path: `{path}` doesn't contains valid identifier."
        )

    def _load_fileset_from_server(self, identifier: NameIdentifier) -> Fileset:
        """Load the fileset from the server.
        If the fileset is not found on the server, an `NoSuchFilesetException` exception will be raised.
        :param identifier: The fileset identifier
        :return The fileset
        """
        catalog: Catalog = self._client.load_catalog(identifier.name())
        return catalog.as_fileset_catalog().load_fileset(identifier)

    def _get_actual_path_by_ident(
        self,
        identifier: NameIdentifier,
        fileset: Fileset,
        fs: AbstractFileSystem,
        storage_type: StorageType,
        virtual_path: str,
    ):
        """Get the actual path by the virtual path and the fileset.
        :param identifier: The fileset identifier
        :param fileset: The fileset
        :param fs: The file system corresponding to the fileset storage location
        :param storage_type: The storage type of the fileset storage location
        :param virtual_path: The virtual fileset path
        :return The actual path.
        """
        virtual_location = self._get_virtual_location(identifier)
        storage_location = fileset.storage_location()
        if self._check_mount_single_file(fileset, fs, storage_type):
            if virtual_path != virtual_location:
                raise GravitinoRuntimeException(
                    f"Path: {virtual_path} should be same with the virtual location: {virtual_location}"
                    " when the fileset only mounts a single file."
                )
            return storage_location
        return virtual_path.replace(virtual_location, storage_location, 1)

    @staticmethod
    def _get_virtual_location(identifier: NameIdentifier):
        """Get the virtual location of the fileset.
        :param identifier: The name identifier of the fileset
        :return The virtual location.
        """
        return (
            f"fileset/{identifier.namespace().level(1)}"
            f"/{identifier.namespace().level(2)}"
            f"/{identifier.name()}"
        )

    def _check_mount_single_file(
        self, fileset: Fileset, fs: AbstractFileSystem, storage_type: StorageType
    ):
        """Check if the fileset is mounted a single file.
        :param fileset: The fileset
        :param fs: The file system corresponding to the fileset storage location
        :param storage_type: The storage type of the fileset storage location
        :return True the fileset is mounted a single file.
        """
        result: Dict = fs.info(
            self._strip_storage_protocol(storage_type, fileset.storage_location())
        )
        return result["type"] == "file"

    @staticmethod
    def _pre_process_path(virtual_path):
        """Pre-process the path.
         We will uniformly process `gvfs://fileset/{catalog}/{schema}/{fileset_name}/xxx`
         into the format of `fileset/{catalog}/{schema}/{fileset_name}/xxx`.
         This is because some implementations of `PyArrow` and `fsspec` can only recognize this format.
        :param virtual_path: The virtual path
        :return The pre-processed path
        """
        if isinstance(virtual_path, PurePosixPath):
            pre_processed_path = virtual_path.as_posix()
        else:
            pre_processed_path = virtual_path
        gvfs_prefix = f"{PROTOCOL_NAME}://"
        if pre_processed_path.startswith(gvfs_prefix):
            pre_processed_path = pre_processed_path[len(gvfs_prefix) :]
        if not pre_processed_path.startswith("fileset/"):
            raise GravitinoRuntimeException(
                f"Invalid path:`{pre_processed_path}`. Expected path to start with `fileset/`."
                " Example: fileset/{fileset_catalog}/{schema}/{fileset_name}/{sub_path}."
            )
        return pre_processed_path

    @staticmethod
    def _strip_storage_protocol(storage_type: StorageType, path: str):
        """Strip the storage protocol from the path.
          Before passing the path to the underlying file system for processing,
           pre-process the protocol information in the path.
          Some file systems require special processing.
          For HDFS, we can pass the path like 'hdfs://{host}:{port}/xxx'.
          For Local, we can pass the path like '/tmp/xxx'.
        :param storage_type: The storage type
        :param path: The path
        :return: The stripped path
        """
        if storage_type == StorageType.HDFS:
            return path
        if storage_type == StorageType.LOCAL:
            return path[len(f"{StorageType.LOCAL.value}:") :]
        raise GravitinoRuntimeException(
            f"Storage type:{storage_type} doesn't support now."
        )


fsspec.register_implementation(PROTOCOL_NAME, GravitinoVirtualFileSystem)



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from typing import ClassVar

from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin, config

from gravitino.exceptions.illegal_name_identifier_exception import (
    IllegalNameIdentifierException,
)
from gravitino.namespace import Namespace


@dataclass
class NameIdentifier(DataClassJsonMixin):
    """A name identifier is a sequence of names separated by dots. It's used to identify a metalake, a
    catalog, a schema or a table. For example, "metalake1" can represent a metalake,
    "metalake1.catalog1" can represent a catalog, "metalake1.catalog1.schema1" can represent a
    schema.
    """

    _DOT: ClassVar[str] = "."

    _name: str = field(metadata=config(field_name="name"))
    _namespace: Namespace = field(metadata=config(field_name="namespace"))

    @classmethod
    def builder(cls, namespace: Namespace, name: str):
        return NameIdentifier(_namespace=namespace, _name=name)

    def namespace(self):
        return self._namespace

    def name(self):
        return self._name

    @staticmethod
    def of(*names: str) -> "NameIdentifier":
        """Create the NameIdentifier with the given levels of names.

        Args:
            names The names of the identifier

        Returns:
            The created NameIdentifier
        """

        NameIdentifier.check(
            names is not None, "Cannot create a NameIdentifier with null names"
        )
        NameIdentifier.check(
            len(names) > 0, "Cannot create a NameIdentifier with no names"
        )

        return NameIdentifier.builder(Namespace.of(*names[:-1]), names[-1])

    @staticmethod
    def of_namespace(namespace: Namespace, name: str) -> "NameIdentifier":
        """Create the NameIdentifier with the given Namespace and name.

        Args:
            namespace: The namespace of the identifier
            name: The name of the identifier

        Returns:
            The created NameIdentifier
        """
        return NameIdentifier.builder(namespace, name)

    @staticmethod
    def of_metalake(metalake: str) -> "NameIdentifier":
        """Create the metalake NameIdentifier with the given name.

        Args:
            metalake: The metalake name

        Returns:
            The created metalake NameIdentifier
        """
        return NameIdentifier.of(metalake)

    @staticmethod
    def of_catalog(metalake: str, catalog: str) -> "NameIdentifier":
        """Create the catalog NameIdentifier with the given metalake and catalog name.

        Args:
            metalake: The metalake name
            catalog: The catalog name

        Returns:
            The created catalog NameIdentifier
        """
        return NameIdentifier.of(metalake, catalog)

    @staticmethod
    def of_schema(metalake: str, catalog: str, schema: str) -> "NameIdentifier":
        """Create the schema NameIdentifier with the given metalake, catalog and schema name.

        Args:
            metalake: The metalake name
            catalog: The catalog name
            schema: The schema name

        Returns:
            The created schema NameIdentifier
        """
        return NameIdentifier.of(metalake, catalog, schema)

    @staticmethod
    def of_table(
        metalake: str, catalog: str, schema: str, table: str
    ) -> "NameIdentifier":
        """Create the table NameIdentifier with the given metalake, catalog, schema and table name.

        Args:
            metalake: The metalake name
            catalog: The catalog name
            schema: The schema name
            table: The table name

        Returns:
            The created table NameIdentifier
        """
        return NameIdentifier.of(metalake, catalog, schema, table)

    @staticmethod
    def of_fileset(
        metalake: str, catalog: str, schema: str, fileset: str
    ) -> "NameIdentifier":
        """Create the fileset NameIdentifier with the given metalake, catalog, schema and fileset name.

        Args:
            metalake: The metalake name
            catalog: The catalog name
            schema: The schema name
            fileset: The fileset name

        Returns:
            The created fileset NameIdentifier
        """
        return NameIdentifier.of(metalake, catalog, schema, fileset)

    @staticmethod
    def of_topic(
        metalake: str, catalog: str, schema: str, topic: str
    ) -> "NameIdentifier":
        """Create the topic NameIdentifier with the given metalake, catalog, schema and topic
        name.

        Args:
            metalake: The metalake name
            catalog: The catalog name
            schema: The schema name
            topic: The topic name

        Returns:
            The created topic NameIdentifier
        """
        return NameIdentifier.of(metalake, catalog, schema, topic)

    @staticmethod
    def check_metalake(ident: "NameIdentifier") -> None:
        """Check the given NameIdentifier is a metalake identifier. Throw an {@link
        IllegalNameIdentifierException} if it's not.

        Args:
            ident: The metalake NameIdentifier to check.
        """
        NameIdentifier.check(ident is not None, "Metalake identifier must not be null")
        Namespace.check_metalake(ident.namespace())

    @staticmethod
    def check_catalog(ident: "NameIdentifier") -> None:
        """Check the given NameIdentifier is a catalog identifier. Throw an {@link
        IllegalNameIdentifierException} if it's not.

        Args:
            ident: The catalog NameIdentifier to check.
        """
        NameIdentifier.check(ident is not None, "Catalog identifier must not be null")
        Namespace.check_catalog(ident.namespace())

    @staticmethod
    def check_schema(ident: "NameIdentifier") -> None:
        """Check the given NameIdentifier is a schema identifier. Throw an {@link
        IllegalNameIdentifierException} if it's not.

        Args:
            ident: The schema NameIdentifier to check.
        """
        NameIdentifier.check(ident is not None, "Schema identifier must not be null")
        Namespace.check_schema(ident.namespace())

    @staticmethod
    def check_table(ident: "NameIdentifier") -> None:
        """Check the given NameIdentifier is a table identifier. Throw an {@link
        IllegalNameIdentifierException} if it's not.

        Args:
            ident: The table NameIdentifier to check.
        """
        NameIdentifier.check(ident is not None, "Table identifier must not be null")
        Namespace.check_table(ident.namespace())

    @staticmethod
    def check_fileset(ident: "NameIdentifier") -> None:
        """Check the given NameIdentifier is a fileset identifier. Throw an {@link
        IllegalNameIdentifierException} if it's not.

        Args:
            ident: The fileset NameIdentifier to check.
        """
        NameIdentifier.check(ident is not None, "Fileset identifier must not be null")
        Namespace.check_fileset(ident.namespace())

    @staticmethod
    def check_topic(ident: "NameIdentifier") -> None:
        """Check the given NameIdentifier is a topic identifier. Throw an {@link
        IllegalNameIdentifierException} if it's not.

        Args:
            ident: The topic NameIdentifier to check.
        """
        NameIdentifier.check(ident is not None, "Topic identifier must not be null")
        Namespace.check_topic(ident.namespace())

    @staticmethod
    def parse(identifier: str) -> "NameIdentifier":
        """Create a NameIdentifier from the given identifier string.

        Args:
            identifier: The identifier string

        Returns:
            The created NameIdentifier
        """
        NameIdentifier.check(
            identifier is not None and identifier != "",
            "Cannot parse a null or empty identifier",
        )

        parts = identifier.split(NameIdentifier._DOT)
        return NameIdentifier.of(*parts)

    def has_namespace(self):
        """Check if the NameIdentifier has a namespace.

        Returns:
            True if the NameIdentifier has a namespace, false otherwise.
        """
        return not self.namespace().is_empty()

    def get_namespace(self):
        """Get the namespace of the NameIdentifier.

        Returns:
            The namespace of the NameIdentifier.
        """
        return self._namespace

    def get_name(self):
        """Get the name of the NameIdentifier.

        Returns:
            The name of the NameIdentifier.
        """
        return self._name

    def __eq__(self, other):
        if not isinstance(other, NameIdentifier):
            return False
        return self._namespace == other._namespace and self._name == other._name

    def __hash__(self):
        return hash((self._namespace, self._name))

    def __str__(self):
        if self.has_namespace():
            return str(self._namespace) + "." + self._name
        return self._name

    @staticmethod
    def check(condition, message, *args):
        """Check the given condition is true. Throw an {@link IllegalNameIdentifierException} if it's not.

        Args:
            condition: The condition to check.
            message: The message to throw.
            args: The arguments to the message.
        """
        if not condition:
            raise IllegalNameIdentifierException(message.format(*args))



"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

from gravitino.api.catalog import Catalog
from gravitino.api.schema import Schema
from gravitino.api.fileset import Fileset
from gravitino.api.fileset_change import FilesetChange
from gravitino.api.metalake_change import MetalakeChange
from gravitino.api.schema_change import SchemaChange
from gravitino.client.gravitino_client import GravitinoClient
from gravitino.client.gravitino_admin_client import GravitinoAdminClient
from gravitino.client.gravitino_metalake import GravitinoMetalake
from gravitino.name_identifier import NameIdentifier
from gravitino.filesystem import gvfs
